/* Brent.c */

#include <stdio.h>
#include <math.h>

#define Nb        200

#define minimum(a, b) ((a) <= (b) ? (a) : (b))
#define maximum(a, b) ((a) >= (b) ? (a) : (b))

/* Common Block Declarations */

struct {
  int nfcall;
} refnum;

#define refnums refnum

int roots(double *x, int j, int color);
int roots2(double xinit, double yinit, int j, int color);
int roots3(double *x, int j, int color, double rad, double *v);
int roots4(double *x, int j, int color, double rad, double *v);
int root_finder(double swept, double *x, int color);
int root_finder2(double swept, double *x, int color, double rad);
double signum(double *a, double *b);
int funcr(int n, double *x, double *fvec, int iflag);
int funcr2(int n, double *x, double *fvec, int iflag, double *v);
int brent(int (*fcn)(), int n, double *x, double *fvec, double tol,
           int info, double *wa, int lwa);
int brentm(int (*fcn)(), int n, double *x, double *fvec, double ftol,
           double xtol, int maxfev, int mopt, int info, int nfev,
           double *q, int ldq, double *sigma, double *wa1, double *wa2);

// extern double pars[Ne];

/* Main program */

int root_finder(double swept, double *x, int color)
{
  int i, ir = 0;

  VEC x_init;
  
  double dxi = swept;
  double dxf, dyf;
  double xf  = BOUND;  

  switch(SWEPT) {
  case 1:
    for(dxf = - xf; dxf <= xf; dxf += dxi) {      
      x_init[0] = dxf; 
      x_init[1] = 0.0;
    
      roots(x_init, ir, color);

      ir += 1;
    }

    /* for(dyf = - xf, dyf != 0.0; dyf <= xf; dyf += dxi) { */

    for(dyf = - xf; dyf <= xf; dyf += dxi) { 
      x_init[0] = 0.0; 
      x_init[1] = dyf;
    
      roots(x_init, ir, color);

      ir += 1;          
    }

    break;
  case 2:
    for(dxf = - xf; dxf <= xf; dxf += dxi) {
      for(dyf = - xf; dyf <= xf; dyf += dxi) {    
        x_init[0] = dxf; 
        x_init[1] = dyf;
    
        roots(x_init, ir, color);        

        /*
          
        printf("x[%d] = %g, y[%d] = %g\n", ir, root_x[ir], ir, root_y[ir]); 
        printf("xl[%d] = %g, yl[%d] = %g\n", ir, Y_x[ir], ir, Y_y[ir]);

        **/
        
        ir += 1;
      }     
    }  

    break;
  case 3:
    for(i = 0; i < Number_of_roots; i++) {    
      roots2(Xr[i][0], Xr[i][1], i, 4);
    }

    break;
  default:
    break;
  }

  return(0);
}

int root_finder2(double swept, double *x, int color, double rad)
{
  int ir = 0;

  VEC x_init;
  
  double dxi = swept, phi, dphi = 0.01;
  double dxf, dyf, r = parms[P_R]; 
  double xf  = BOUND, Npi = 2.0 * M_PI;  

  z1 = source_position.x;
  z2 = source_position.y;

  for(phi = 0; phi < Npi; phi = phi + dphi) {
    pars[0] = r * cos(phi) + z1;
    pars[1] = r * sin(phi) + z2;
 
    for(dxf = - xf; dxf <= xf; dxf += dxi) {
      for(dyf = - xf; dyf <= xf; dyf += dxi) {    
	x_init[0] = dxf; 
	x_init[1] = dyf;
	
	roots4(x_init, ir, color, rad, pars);        
	
	ir += 1;
      }     
    }  
  }
  
  return(0);
}

int roots(double *x, int j, int color)
{
  /* Local variables */

  static double fvec[Ne];
  static int info, nfev;
  static int i;
  static double fnorm1, fnorm2, wa[Nb];
  static int lwa;
  static double tol, sqroot;

  /*     LOGICAL OUTPUT UNIT IS ASSUMED TO BE NUMBER 6. */

  lwa = Nb;

  tol = 1.0e-10;

  /*     STARTING VALUES. */

  for(i = 0; i < Ne; i++) {
    fvec[i] = 0.0;
  }

  for(i = 0; i < Nb; i++) {
    wa[i] = 0.0;
  }
  
  /*     INITIAL MAX-NORM OF THE RESIDUALS. */

  fnorm1 = 0.0;

  for(i = 0; i < Ne; i++) {
    funcr(Ne, x, fvec, i); 

    fnorm1 = maximum(fnorm1, fabs(fvec[i]));
  }

  refnums.nfcall = 0;

  brent(funcr, Ne, x, fvec, tol, info, wa, lwa);
  
  nfev = refnums.nfcall / Ne;

  /*     FINAL MAX-NORM OF THE RESIDUALS. */

  fnorm2 = 0.0;

  for(i = 0; i < Ne; i++) {
    funcr(Ne, x, fvec, i);

    fnorm2 = maximum(fnorm2, fabs(fvec[i]));
  }

  printf("NUMBER OF FUNCTION EVALUATIONS = %d \n", nfev);
  printf("EXIT PARAMETER = %d \n", info);
  printf("FINAL APPROXIMATE SOLUTION \n");

  sqroot = sqrt(fvec[0] * fvec[0] + fvec[1] * fvec[1]);
    
  if(sqroot <= tol) {
    draw_point(x[0], x[1], 5.0, color);

    for(i = 0; i < Ne; i++) {
      printf("x[%d] = % g \n", i, x[i]);
      printf("fvec[%d] = %g \n", i, fvec[i]);
    }

    root_x[j] = x[0];
    root_y[j] = x[1];
  }  
  
  return 0;
}

int roots2(double xinit, double yinit, int j, int color)
{
  /* Local variables */

  static double fvec[Ne];
  static int info, nfev;
  static int i;
  static double fnorm1, fnorm2, wa[Nb];
  static int lwa;
  static double tol, sqroot, x[Ne];

  /*     LOGICAL OUTPUT UNIT IS ASSUMED TO BE NUMBER 6. */

  lwa = Nb;

  tol = 1.0e-10;

  /*     STARTING VALUES. */

  for(i = 0; i < Ne; i++) {
    fvec[i] = 0.0;
  }

  for(i = 0; i < Nb; i++) {
    wa[i] = 0.0;
  }
  
  /*     INITIAL MAX-NORM OF THE RESIDUALS. */

  fnorm1 = 0.0;

  x[0] = xinit;
  x[1] = yinit;
    
  for(i = 0; i < Ne; i++) {
    funcr(Ne, x, fvec, i); 

    fnorm1 = maximum(fnorm1, fabs(fvec[i]));
  }

  refnums.nfcall = 0;

  brent(funcr, Ne, x, fvec, tol, info, wa, lwa);
  
  nfev = refnums.nfcall / Ne;

  /*     FINAL MAX-NORM OF THE RESIDUALS. */

  fnorm2 = 0.0;

  for(i = 0; i < Ne; i++) {
    funcr(Ne, x, fvec, i);

    fnorm2 = maximum(fnorm2, fabs(fvec[i]));
  }

  printf("NUMBER OF FUNCTION EVALUATIONS = %d \n", nfev);
  printf("EXIT PARAMETER = %d \n", info);
  printf("FINAL APPROXIMATE SOLUTION \n");

  sqroot = sqrt(fvec[0] * fvec[0] + fvec[1] * fvec[1]);
    
  if(sqroot <= tol) {
    draw_point(x[0], x[1], 5.0, color);

    for(i = 0; i < Ne; i++) {
      printf("x[%d] = % g \n", i, x[i]);
      printf("fvec[%d] = %g \n", i, fvec[i]);
    }

    root_x[j] = x[0];
    root_y[j] = x[1];
  }  
  
  return 0;
}

int roots3(double *x, int j, int color, double rad, double *v)
{
  /* Local variables */

  static double fvec[Ne];
  static int info = 0, nfev;
  static int i;
  static double fnorm1, fnorm2, wa[Nb];
  static int lwa;
  static double tol, sqroot;

  /*     LOGICAL OUTPUT UNIT IS ASSUMED TO BE NUMBER 6. */

  lwa = Nb;

  tol = 1.0e-10;

  /*     STARTING VALUES. */

  for(i = 0; i < Ne; i++) {
    fvec[i] = 0.0;
  }

  for(i = 0; i < Nb; i++) {
    wa[i] = 0.0;
  }
  
  /*     INITIAL MAX-NORM OF THE RESIDUALS. */

  fnorm1 = 0.0;

  for(i = 0; i < Ne; i++) {
    funcr2(Ne, x, fvec, i, v); 

    fnorm1 = maximum(fnorm1, fabs(fvec[i]));
  }

  refnums.nfcall = 0;

  brent(funcr2, Ne, x, fvec, tol, info, wa, lwa);
  
  nfev = refnums.nfcall / Ne;

  /*     FINAL MAX-NORM OF THE RESIDUALS. */

  fnorm2 = 0.0;

  for(i = 0; i < Ne; i++) {
    funcr2(Ne, x, fvec, i, v);

    fnorm2 = maximum(fnorm2, fabs(fvec[i]));
  }

  /*
    
  printf("NUMBER OF FUNCTION EVALUATIONS = %d \n", nfev);
  printf("EXIT PARAMETER = %d \n", info);
  printf("FINAL APPROXIMATE SOLUTION \n");

  **/
  
  sqroot = sqrt(fvec[0] * fvec[0] + fvec[1] * fvec[1]);

  if(sqroot <= tol) {
    draw_point(x[0], x[1], rad, color);

    /*
      
    for(i = 0; i < Ne; i++) {
      printf("x[%d] = % g \n", i, x[i]);
      printf("fvec[%d] = %g \n", i, fvec[i]);
    }

    **/
    
    root_x[j] = x[0];
    root_y[j] = x[1];
  }  
  
  return 0;
}

int roots4(double *x, int j, int color, double rad, double *v)
{
  /* Local variables */

  static double fvec[Ne];
  static int info = 0, nfev;
  static int i;
  static double fnorm1, fnorm2, wa[Nb];
  static int lwa;
  static double tol, sqroot;

  /*     LOGICAL OUTPUT UNIT IS ASSUMED TO BE NUMBER 6. */

  lwa = Nb;

  tol = 1.0e-10;

  /*     STARTING VALUES. */

  for(i = 0; i < Ne; i++) {
    fvec[i] = 0.0;
  }

  for(i = 0; i < Nb; i++) {
    wa[i] = 0.0;
  }
  
  /*     INITIAL MAX-NORM OF THE RESIDUALS. */

  fnorm1 = 0.0;

  for(i = 0; i < Ne; i++) {
    funcr2(Ne, x, fvec, i, v); 

    fnorm1 = maximum(fnorm1, fabs(fvec[i]));
  }

  refnums.nfcall = 0;

  brent(funcr2, Ne, x, fvec, tol, info, wa, lwa);
  
  nfev = refnums.nfcall / Ne;

  /*     FINAL MAX-NORM OF THE RESIDUALS. */

  fnorm2 = 0.0;

  for(i = 0; i < Ne; i++) {
    funcr2(Ne, x, fvec, i, v);

    fnorm2 = maximum(fnorm2, fabs(fvec[i]));
  }

  /*
    
  printf("NUMBER OF FUNCTION EVALUATIONS = %d \n", nfev);
  printf("EXIT PARAMETER = %d \n", info);
  printf("FINAL APPROXIMATE SOLUTION \n");

  **/
  
  sqroot = sqrt(fvec[0] * fvec[0] + fvec[1] * fvec[1]);

  if(sqroot <= tol) {
    for(i = 0; i < Ne; i++) {
      printf("x[%d] = % g \n", i, x[i]);
      printf("fvec[%d] = %g \n", i, fvec[i]);
    }
    
    draw_point(x[0], x[1], rad, color);
  }
  
  return 0;
}

double signum(double *a, double *b)
{
  double x; 

  x = ((*a) >= 0 ? (*a) : - (*a));

  return((*b) >= 0 ? x : - x);
}

int funcr(int n, double *x, double *fvec, int iflag)
{
  double func1, func2, func3, funchb, funcnis, funcrl, funcs, functs;
  double poly1, poly2, poly3, poly4, poly5, poly6, poly7;
  double dipolm1, dipolm2, quadrupolm1, quadrupolm2; 
  double rbeta, s2, rs12, radius_E4;  
  double B1, B2, D2, yp1, yp2, rd2, xd1, xd2, Mkaprp, Mkapto;  
  double alpha0 = 0.0, alpha1 = 0.0;
  double hubble_factor =  2.0;    
  double ring_factor   = 10.0;
  double Epsilon1 = 1.0 - parms[P_EPSILON];
  double Epsilon2 = 1.0 + parms[P_EPSILON];
  double delta  = 1.0 - parms[P_ALPHA];
  
  // double delta2 = 2.0 * parms[P_ALPHA];  
  // double xbar0 =   (x[0] - xl) * cs + (x[1] - yl) * ss;
  // double xbar1 = - (x[0] - xl) * ss + (x[1] - yl) * cs;  

  double xbar0 = 0.0, xbar1 = 0.0;
  
  // double radius_E2 = (xbar0 * xbar0 + xbar1 * xbar1) / R_E2;
  // double radius_E  = sqrt(radius_E2);

  double radius_E, radius_E2;
  double alpha_x, alpha_y;

  xl = 0.0;
  yl = 0.0;
  
  xbar0 =   (x[0] - xl) * cs + (x[1] - yl) * ss;
  xbar1 = - (x[0] - xl) * ss + (x[1] - yl) * cs;  

  radius_E2 = (xbar0 * xbar0 + xbar1 * xbar1) / R_E2;
  radius_E  = sqrt(radius_E2);
 
  switch(current_model) {
  case M_MODEL_MENU:
    break;    
  case M_CHANG_REFSDAL:
    alpha0 = xbar0 / radius_E2;
    alpha1 = xbar1 / radius_E2;
  
    break;
  case M_SECOND:
    
    break;      
  case M_SING_ISOT_SPHERE:
    rbeta = pow(radius_E, parms[P_EXP] - 1.0);
    
    alpha0 = rbeta * xbar0;
    alpha1 = rbeta * xbar1;  
  
    break;      
  case M_DE_VAUCOULEUR:
    poly1 = 720.0 / tau8 + (720.0 * pow(radius_E, (1.0 / 4.0))) / tau7;
    poly2 = poly1 + pow(radius_E, (3.0 / 2.0)) / tau2;
    poly3 = poly2 + (6.0 * pow(radius_E, (5.0 / 4.0))) / tau3;
    poly4 = poly3 + (30.0 * radius_E) / tau4;
    poly5 = poly4 + (120.0 * pow(radius_E, (3.0 / 4.0))) / tau5;
    poly6 = poly5 + (360.0 * sqrt(radius_E)) / tau6;
    poly7 = 7.0 * poly6 + pow(radius_E, (7.0 / 4.0)) / tau;

    func1 = poly7 * exp(- tau * (pow(radius_E, (1.0 / 4.0)) - 1.0));
    func2 = func1 - (5040.0 * Ftau) / tau8;

    alpha0 = - (8.0 * parms[P_KAPPA] * xbar0 * func2) / radius_E2;
    alpha1 = - (8.0 * parms[P_KAPPA] * xbar1 * func2) / radius_E2;
  
    break;      
  case M_KING: 
    /* log1p(x) = log(1.0 + x) */
  
    alpha0 = (parms[P_KAPPA] * xbar0 * log1p(radius_E2)) / radius_E2;
    alpha1 = (parms[P_KAPPA] * xbar1 * log1p(radius_E2)) / radius_E2;
  
    break;      
  case M_HUBBLE:
    funchb = log1p(radius_E) - radius_E / (1.0 + radius_E);

    alpha0 = (hubble_factor * parms[P_KAPPA] * xbar0 * funchb) / radius_E2;
    alpha1 = (hubble_factor * parms[P_KAPPA] * xbar1 * funchb) / radius_E2;  

    break;      
  case M_SPIRAL:
    funcs = 2.0 * parms[P_KAPPA] * (1.0 - exp(- parms[P_ALPHA2] * radius_E) 
				    * (parms[P_ALPHA2] * radius_E + 1.0));
    
    alpha0 = (funcs * xbar0) / (parms[P_ALPHA2] * parms[P_ALPHA2] * radius_E2);
    alpha1 = (funcs * xbar1) / (parms[P_ALPHA2] * parms[P_ALPHA2] * radius_E2);
    
    break;      
  case M_TRUNC_KING:
    /* hypot(x, y) = sqrt(x * x + y * y) */

    func1 = log1p(radius_E2) + radius_E2 / (1.0 + xc2); 
    func2 = (4.0 * (hypot(1.0, radius_E) - 1.0)) / hypot(1.0, xc);
    func3 = func1 - func2;
  
    alpha0 = (parms[P_KAPPA2] * xbar0 * func3) / radius_E2; 
    alpha1 = (parms[P_KAPPA2] * xbar1 * func3) / radius_E2;

    break;      
  case M_NSING_ISOT_SPHERE:
     s2 = xc2 + radius_E2;
    
     funcnis = sqrt(s2) - xc;

     alpha0 = (parms[P_KAPPA] * funcnis * xbar0) / radius_E2;
     alpha1 = (parms[P_KAPPA] * funcnis * xbar1) / radius_E2;
  
     break;      
  case M_TRANSP_SPHERE:
    if(parms[P_C] > parms[P_E] * radius_E){
      s2 = 1.0 - radius_E2 * (R_E2 / R_C2);

      functs = 1.0 - pow(s2, (3.0 / 2.0));

      alpha0 = (functs * xbar0) / radius_E2;
      alpha1 = (functs * xbar1) / radius_E2;    
    } else {
      alpha0 = xbar0 / radius_E2;
      alpha1 = xbar1 / radius_E2; 
    }   

    break;
  case M_ELLIPTICAL:
    s2 = xc2 + (Epsilon1 * xbar0 * xbar0 + Epsilon2 * xbar1 * xbar1) / R_E2;

    /*

    alpha0 = (parms[P_KAPPA] * Epsilon1 
	      * xbar0) / (pow(s2, delta) * pow(xc, delta2));
    alpha1 = (parms[P_KAPPA] * Epsilon2 
	      * xbar1) / (pow(s2, delta) * pow(xc, delta2));

    **/

    alpha0 = (parms[P_KAPPA] * Epsilon1 * xbar0) / pow(s2, delta);
    alpha1 = (parms[P_KAPPA] * Epsilon2 * xbar1) / pow(s2, delta);

    break;      
  case M_MULTIPOLE:
    dipolm1 =   R_E3 * (parms[P_DX_QP] * (xbar1 * xbar1 - xbar0 * xbar0)
                        - 2.0 * parms[P_DY_QP] * xbar0 *
                        xbar1) / pow((xbar0 * xbar0 + xbar1 * xbar1), 2.0);
    
    dipolm2 = - R_E3 * (parms[P_DY_QP] * (xbar1 * xbar1 - xbar0 * xbar0)
                        + 2.0 * parms[P_DX_QP] * xbar0 *
                        xbar1) / pow((xbar0 * xbar0 + xbar1 * xbar1), 2.0);

    quadrupolm1 = 2.0 * R_E4 * (parms[P_QXX] * xbar0 *
                                (3.0 * xbar1 * xbar1 - xbar0 * xbar0)
                                + parms[P_QXY] * xbar1 *
                                (xbar1 * xbar1 - 3.0 * xbar0 * xbar0)) /
      pow((xbar0 * xbar0 + xbar1 * xbar1), 3.0);

    quadrupolm2 = 2.0 * R_E4 * (parms[P_QYY] * xbar1 *
                                (3.0 * xbar0 * xbar0 - xbar1 * xbar1)
                                + parms[P_QXY] * xbar0 *
                                (xbar0 * xbar0 - 3.0 * xbar1 * xbar1)) /
      pow((xbar0 * xbar0 + xbar1 * xbar1), 3.0);
    
    alpha0 = (parms[P_KAPPA0] * xbar0) / radius_E2 + dipolm1 + quadrupolm1; 
    alpha1 = (parms[P_KAPPA0] * xbar1) / radius_E2 + dipolm2 + quadrupolm2;

    break;
  case M_ROTATION_LENS:
    funcrl = (parms[P_SX] * xbar1 - parms[P_SY] * xbar0) / parms[P_E];

    radius_E2 = (xbar0 * xbar0 + xbar1 * xbar1) / R_E2;
    radius_E4 = radius_E2 * radius_E2;

    alpha0 = xbar0 / radius_E2 + (2.0 * funcrl * xbar0) / radius_E4
      + parms[P_SY] / (parms[P_E] * radius_E2);
    alpha1 = xbar1 / radius_E2 + (2.0 * funcrl * xbar1) / radius_E4
      - parms[P_SX] / (parms[P_E] * radius_E2);
  
    break;
  case M_DOUBLE_LENS:
    rs12 = ((xbar0 - parms[P_X_1]) * (xbar0 - parms[P_X_1]) + 
            (xbar1 - parms[P_Y_1]) * (xbar1 - parms[P_Y_1])) / R_E2;

    B1 = 1.0 - parms[P_SIGMA2] - parms[P_GAMMA2] * cos(parms[P_PHI2]);
    B2 = 1.0 - parms[P_SIGMA2] + parms[P_GAMMA2] * cos(parms[P_PHI2]);
    D2 =                       - parms[P_GAMMA2] * sin(parms[P_PHI2]);

    yp1 = B1 * xbar0 + D2 * xbar1;
    yp2 = B2 * xbar1 + D2 * xbar0; 
  
    xd1 = yp1 - parms[P_M_1] * parms[P_BETA] * (xbar0 - parms[P_X_1]) / rs12; 
    xd2 = yp2 - parms[P_M_1] * parms[P_BETA] * (xbar1 - parms[P_Y_1]) / rs12; 

    rd2 = ((xd1 - parms[P_X_2]) * (xd1 - parms[P_X_2]) + 
                  (xd2 - parms[P_Y_2]) * (xd2 - parms[P_Y_2])) / R_E2;

    alpha0 = parms[P_M_1] * (xbar0 - parms[P_X_1]) / rs12
      + parms[P_M_2] * (xd1 - parms[P_X_2]) / rd2;  
    alpha1 = parms[P_M_1] * (xbar1 - parms[P_Y_1]) / rs12
      + parms[P_M_2] * (xd2 - parms[P_Y_2]) / rd2;          
    
    break;
  case M_UNIFORM_RING:
    Mkaprp = ring_factor * parms[P_KAPPA] * (radius_E2 - Ri_E2);
    Mkapto = ring_factor * parms[P_KAPPA] * (Ro_E2 - Ri_E2);

    if(radius_E >= Ro_E / parms[P_E]) {
      alpha0 = (Mkapto * xbar0) / radius_E2;
      alpha1 = (Mkapto * xbar1) / radius_E2; 	     
    } else {
      if(radius_E >= Ri_E / parms[P_E]) {
        alpha0 = (Mkaprp * xbar0) / radius_E2;
        alpha1 = (Mkaprp * xbar1) / radius_E2; 	
      } else {
        alpha0 = 0.0;
        alpha1 = 0.0; 
      }   
    }
          
    break;
  default:
    break;
  }
  
  sdist2 = ((xbar0 - dist_1) * (xbar0 - dist_1) 
	    + (xbar1 - dist_2) * (xbar1 - dist_2)) / (parms[P_E] * parms[P_E]);
  
  if(black_hole == 1) {
    alpha_x = (bh_beta * (xbar0 - dist_1)) / sdist2;
    alpha_y = (bh_beta * (xbar1 - dist_2)) / sdist2; 
  } else {
    alpha_x = 0.0;
    alpha_y = 0.0; 
  }

  fvec[0] = Y[0] - A1 * xbar0 - D * xbar1 + alpha0 + alpha_x;
  fvec[1] = Y[1] - A2 * xbar1 - D * xbar0 + alpha1 + alpha_y;

  /*

  fvec[0] = pars[0] - A1 * xbar0 - D * xbar1 + alpha0;
  fvec[1] = pars[1] - A2 * xbar1 - D * xbar0 + alpha1;

  **/
  
  ++refnums.nfcall;

  return(0);
}

int funcr2(int n, double *x, double *fvec, int iflag, double *v)
{
  double func1, func2, func3, funchb, funcnis, funcrl, funcs, functs;
  double poly1, poly2, poly3, poly4, poly5, poly6, poly7;
  double dipolm1, dipolm2, quadrupolm1, quadrupolm2; 
  double alpha0 = 0.0, alpha1 = 0.0, rbeta, s2, rs12, radius_E4;  
  double B1, B2, D2, yp1, yp2, rd2, xd1, xd2, Mkaprp, Mkapto;  
  double hubble_factor =  2.0;    
  double ring_factor   = 10.0;
  double Epsilon1 = 1.0 - parms[P_EPSILON];
  double Epsilon2 = 1.0 + parms[P_EPSILON];
  double delta  = parms[P_ALPHA] - 1.0;
  
  // double delta2 = 2.0 * parms[P_ALPHA];  
  
  // double xbar0 =   (x[0] - xl) * cs + (x[1] - yl) * ss;
  // double xbar1 = - (x[0] - xl) * ss + (x[1] - yl) * cs;  

  double xbar0 = 0.0, xbar1 = 0.0;

  // double radius_E2 = (xbar0 * xbar0 + xbar1 * xbar1) / R_E2;
  // double radius_E  = sqrt(radius_E2);

  double radius_E, radius_E2;
  double alpha_x, alpha_y;
  
  // pars[0] = v[0];
  // pars[1] = v[1];

  xl = 0.0;
  yl = 0.0;
  
  xbar0 =   (x[0] - xl) * cs + (x[1] - yl) * ss;
  xbar1 = - (x[0] - xl) * ss + (x[1] - yl) * cs;  
  
  radius_E2 = (xbar0 * xbar0 + xbar1 * xbar1) / R_E2;
  radius_E  = sqrt(radius_E2);

  switch(current_model) {
  case M_MODEL_MENU:
    break;
  case M_CHANG_REFSDAL:
    alpha0 = xbar0 / radius_E2;
    alpha1 = xbar1 / radius_E2;

    break;
  case M_SECOND:
    
    break;      
  case M_SING_ISOT_SPHERE:
    rbeta = pow(radius_E2, parms[P_EXP] / 2.0 - 1.0);
    
    alpha0 = rbeta * xbar0;
    alpha1 = rbeta * xbar1;  

    break;      
  case M_DE_VAUCOULEUR:
    poly1 = 720.0 / tau8 + (720.0 * pow(radius_E, (1.0 / 4.0))) / tau7;
    poly2 = poly1 + pow(radius_E, (3.0 / 2.0)) / tau2;
    poly3 = poly2 + (6.0 * pow(radius_E, (5.0 / 4.0))) / tau3;
    poly4 = poly3 + (30.0 * radius_E) / tau4;
    poly5 = poly4 + (120.0 * pow(radius_E, (3.0 / 4.0))) / tau5;
    poly6 = poly5 + (360.0 * sqrt(radius_E)) / tau6;
    poly7 = 7.0 * poly6 + pow(radius_E, (7.0 / 4.0)) / tau;

    func1 = poly7 * exp(- tau * (pow(radius_E, (1.0 / 4.0)) - 1.0));
    func2 = func1 - (5040.0 * Ftau) / tau8;

    alpha0 = - (8.0 * parms[P_KAPPA] * xbar0 * func2) / radius_E2;
    alpha1 = - (8.0 * parms[P_KAPPA] * xbar1 * func2) / radius_E2;

    break;      
  case M_KING: 
    /* log1p(x) = log(1.0 + x) */
  
    alpha0 = (parms[P_KAPPA] * xbar0 * log1p(radius_E2)) / radius_E2;
    alpha1 = (parms[P_KAPPA] * xbar1 * log1p(radius_E2)) / radius_E2;
  
    fvec[0] = pars[0] - A1 * xbar0 - D * xbar1 + alpha0; 
    fvec[1] = pars[1] - A2 * xbar1 - D * xbar0 + alpha1;

    break;      
  case M_HUBBLE:
    funchb = log1p(radius_E) - radius_E / (1.0 + radius_E);

    alpha0 = (hubble_factor * parms[P_KAPPA] * xbar0 * funchb) / radius_E2;
    alpha1 = (hubble_factor * parms[P_KAPPA] * xbar1 * funchb) / radius_E2;  

    break;      
  case M_SPIRAL:
    funcs = 2.0 * parms[P_KAPPA] * (1.0 - exp(- parms[P_ALPHA2] * radius_E)
                                    * (radius_E + 1.0));

    alpha0 = (funcs * xbar0) / (parms[P_ALPHA2] * parms[P_ALPHA2] * radius_E2);
    alpha1 = (funcs * xbar1) / (parms[P_ALPHA2] * parms[P_ALPHA2] * radius_E2);
    
    break;      
  case M_TRUNC_KING:
    /* hypot(x, y) = sqrt(x * x + y * y) */

    func1 = log1p(radius_E2) + radius_E2 / (1.0 + xc2); 
    func2 = (4.0 * (hypot(1.0, radius_E) - 1.0)) / hypot(1.0, xc);
    func3 = func1 - func2;
  
    alpha0 = (parms[P_KAPPA2] * xbar0 * func3) / radius_E2; 
    alpha1 = (parms[P_KAPPA2] * xbar1 * func3) / radius_E2;

    break;      
  case M_NSING_ISOT_SPHERE:
     s2 = xc2 + radius_E2;
    
     funcnis = sqrt(s2) - xc;

     alpha0 = (parms[P_KAPPA] * funcnis * xbar0) / radius_E2;
     alpha1 = (parms[P_KAPPA] * funcnis * xbar1) / radius_E2;
  
     break;      
  case M_TRANSP_SPHERE:
    s2 = 1.0 - radius_E2 / R_C2;

    functs = 1.0 - pow(s2, (3.0 / 2.0));

    if(parms[P_C] > radius_E){
      alpha0 = (functs * xbar0) / (radius_E2 / R_E2);
      alpha1 = (functs * xbar1) / (radius_E2 / R_E2);    
    } else {
      alpha0 = xbar0 / (radius_E2 / R_E2);
      alpha1 = xbar1 / (radius_E2 / R_E2); 
    }   

    break;
  case M_ELLIPTICAL:
    s2 = xc2 + (Epsilon1 * xbar0 * xbar0 + Epsilon2 * xbar1 * xbar1) / R_E2;
  
    /*

    alpha0 = (parms[P_KAPPA] * Epsilon1 
	      * xbar0) / (pow(s2, delta) * pow(xc, delta2));
    alpha1 = (parms[P_KAPPA] * Epsilon2 
	      * xbar1) / (pow(s2, delta) * pow(xc, delta2));

    **/

    alpha0 = (parms[P_KAPPA] * Epsilon1 * xbar0) / pow(s2, delta);
    alpha1 = (parms[P_KAPPA] * Epsilon2 * xbar1) / pow(s2, delta);

    break;      
  case M_MULTIPOLE:
    dipolm1 =   R_E3 * (parms[P_DX_QP] * (xbar1 * xbar1 - xbar0 * xbar0)
                        - 2.0 * parms[P_DY_QP] * xbar0 *
                        xbar1) / pow((xbar0 * xbar0 + xbar1 * xbar1), 2.0);
    
    dipolm2 = - R_E3 * (parms[P_DY_QP] * (xbar1 * xbar1 - xbar0 * xbar0)
                        + 2.0 * parms[P_DX_QP] * xbar0 *
                        xbar1) / pow((xbar0 * xbar0 + xbar1 * xbar1), 2.0);

    quadrupolm1 = 2.0 * R_E4 * (parms[P_QXX] * xbar0 *
                                (3.0 * xbar1 * xbar1 - xbar0 * xbar0)
                                + parms[P_QXY] * xbar1 *
                                (xbar1 * xbar1 - 3.0 * xbar0 * xbar0)) /
      pow((xbar0 * xbar0 + xbar1 * xbar1), 3.0);

    quadrupolm2 = 2.0 * R_E4 * (parms[P_QYY] * xbar1 *
                                (3.0 * xbar0 * xbar0 - xbar1 * xbar1)
                                + parms[P_QXY] * xbar0 *
                                (xbar0 * xbar0 - 3.0 * xbar1 * xbar1)) /
      pow((xbar0 * xbar0 + xbar1 * xbar1), 3.0);
    
    alpha0 = (parms[P_KAPPA0] * xbar0) / radius_E2 + dipolm1 + quadrupolm1; 
    alpha1 = (parms[P_KAPPA0] * xbar1) / radius_E2 + dipolm2 + quadrupolm2;

    break;
  case M_ROTATION_LENS:
    funcrl = (parms[P_SX] * xbar1 - parms[P_SY] * xbar0) / parms[P_E];

    radius_E2 = (xbar0 * xbar0 + xbar1 * xbar1) / R_E2;
    radius_E4 = radius_E2 * radius_E2;

    alpha0 = xbar0 / radius_E2 + (2.0 * funcrl * xbar0) / radius_E4
      + parms[P_SY] / (parms[P_E] * radius_E2);
    alpha1 = xbar1 / radius_E2 + (2.0 * funcrl * xbar1) / radius_E4
      - parms[P_SX] / (parms[P_E] * radius_E2);
  
    break;
  case M_DOUBLE_LENS:
    rs12 = ((xbar0 - parms[P_X_1]) * (xbar0 - parms[P_X_1]) + 
            (xbar1 - parms[P_Y_1]) * (xbar1 - parms[P_Y_1])) / R_E2;

    B1 = 1.0 - parms[P_SIGMA2] - parms[P_GAMMA2] * cos(parms[P_PHI2]);
    B2 = 1.0 - parms[P_SIGMA2] + parms[P_GAMMA2] * cos(parms[P_PHI2]);
    D2 =                       - parms[P_GAMMA2] * sin(parms[P_PHI2]);

    yp1 = B1 * xbar0 + D2 * xbar1;
    yp2 = B2 * xbar1 + D2 * xbar0; 
  
    xd1 = yp1 - parms[P_M_1] * parms[P_BETA] * (xbar0 - parms[P_X_1]) / rs12; 
    xd2 = yp2 - parms[P_M_1] * parms[P_BETA] * (xbar1 - parms[P_Y_1]) / rs12; 

    rd2 = ((xd1 - parms[P_X_2]) * (xd1 - parms[P_X_2]) + 
                  (xd2 - parms[P_Y_2]) * (xd2 - parms[P_Y_2])) / R_E2;

    alpha0 = parms[P_M_1] * (xbar0 - parms[P_X_1]) / rs12
      + parms[P_M_2] * (xd1 - parms[P_X_2]) / rd2;  
    alpha1 = parms[P_M_1] * (xbar1 - parms[P_Y_1]) / rs12
      + parms[P_M_2] * (xd2 - parms[P_Y_2]) / rd2;          
    
    break;
  case M_UNIFORM_RING:
    Mkaprp = ring_factor * parms[P_KAPPA] * (radius_E2 - Ri_E2);
    Mkapto = ring_factor * parms[P_KAPPA] * (Ro_E2 - Ri_E2);

    if(radius_E >= Ro_E) {
      alpha0 = (Mkapto * xbar0) / radius_E2;
      alpha1 = (Mkapto * xbar1) / radius_E2; 	     
    } else {
      if(radius_E >= Ri_E) {
        alpha0 = (Mkaprp * xbar0) / radius_E2;
        alpha1 = (Mkaprp * xbar1) / radius_E2; 	
      } else {
        alpha0 = 0.0;
        alpha1 = 0.0; 
      }   
    }
          
    break;
  default:
    break;
  }

  sdist2 = ((xbar0 - dist_1) * (xbar0 - dist_1) 
	    + (xbar1 - dist_2) * (xbar1 - dist_2)) / (parms[P_E] * parms[P_E]);
  
  if(black_hole == 1) {
    alpha_x = (bh_beta * (xbar0 - dist_1)) / sdist2;
    alpha_y = (bh_beta * (xbar1 - dist_2)) / sdist2; 
  } else {
    alpha_x = 0.0;
    alpha_y = 0.0; 
  }

  fvec[0] = Y[0] - A1 * xbar0 - D * xbar1 + alpha0 + alpha_x;
  fvec[1] = Y[1] - A2 * xbar1 - D * xbar0 + alpha1 + alpha_y;

  /*

  fvec[0] = pars[0] - A1 * xbar0 - D * xbar1 + alpha0;
  fvec[1] = pars[1] - A2 * xbar1 - D * xbar0 + alpha1;

  **/

  ++refnums.nfcall;

  return(0);
}

/*     **********                                                       */

/*     SUBROUTINE BRENT                                                 */

/*     THE PURPOSE OF THIS SUBROUTINE IS TO FIND A ZERO OF              */
/*     A SYSTEM OF N NONLINEAR EQUATIONS IN N VARIABLES BY A            */
/*     METHOD DUE TO R. BRENT. THIS IS DONE BY USING THE                */
/*     MORE GENERAL NONLINEAR EQUATION SOLVER BRENTM.                   */

/*     THE SUBROUTINE STATEMENT IS                                      */

/*       BRENT(FCN, N, X, FVEC, TOL, INFO, WA, LWA)                     */

/*     WHERE                                                            */

/*       FCN IS THE NAME OF THE USER-SUPPLIED SUBROUTINE WHICH          */
/*         CALCULATES COMPONENTS OF THE FUNCTION. FCN SHOULD BE         */
/*         DECLARED IN AN EXTERNAL STATEMENT IN THE USER CALLING        */
/*         PROGRAM, AND SHOULD BE WRITTEN AS FOLLOWS.                   */

/*         FCN(N, X, FVEC, IFLAG)                                       */
/*         INT N, IFLAG                                                 */
/*         DOUBLE X(N), FVEC(N)                                         */
/*         ----------                                                   */
/*         CALCULATE THE IFLAG-TH COMPONENT OF THE FUNCTION             */
/*         AND RETURN THIS VALUE IN FVEC(IFLAG).                        */
/*         ----------                                                   */
/*         RETURN                                                       */
/*         END                                                          */

/*         THE VALUE OF IFLAG SHOULD NOT BE CHANGED BY FCN UNLESS       */
/*         THE USER WANTS TO TERMINATE EXECUTION OF BRENT.              */
/*         IN THIS CASE SET IFLAG TO A NEGATIVE INTEGER.                */

/*       N IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER       */
/*         OF EQUATIONS AND VARIABLES.                                  */

/*       X IS AN ARRAY OF LENGTH N. ON INPUT IT MUST CONTAIN            */
/*         AN INITIAL ESTIMATE OF THE SOLUTION VECTOR. ON OUTPUT X      */
/*         CONTAINS THE FINAL ESTIMATE OF THE SOLUTION VECTOR.          */

/*       FVEC IS AN ARRAY OF LENGTH N. ON OUTPUT IT CONTAINS            */
/*         THE FINAL RESIDUALS.                                         */

/*       TOL IS A NONNEGATIVE INPUT VARIABLE. THE ALGORITHM CONVERGES   */
/*         IF EITHER ALL THE RESIDUALS ARE AT MOST TOL IN MAGNITUDE,    */
/*         OR IF THE ALGORITHM ESTIMATES THAT THE RELATIVE ERROR        */
/*         BETWEEN X AND THE SOLUTION IS AT MOST TOL.                   */

/*       INFO IS AN INTEGER OUTPUT VARIABLE SET AS FOLLOWS. IF          */
/*         THE USER HAS TERMINATED EXECUTION, INFO WILL BE SET TO       */
/*         THE (NEGATIVE) VALUE OF IFLAG. SEE DESCRIPTION OF FCN.       */
/*         OTHERWISE                                                    */

/*         INFO = 0   IMPROPER INPUT PARAMETERS.                        */

/*         INFO = 1   ALL RESIDUALS ARE AT MOST TOL IN MAGNITUDE.       */

/*         INFO = 2   ALGORITHM ESTIMATES THAT THE RELATIVE ERROR       */
/*                    BETWEEN X AND THE SOLUTION IS AT MOST TOL.        */

/*         INFO = 3   CONDITIONS FOR INFO = 1 AND INFO = 2 BOTH HOLD.   */

/*         INFO = 4   NUMBER OF FUNCTION EVALUATIONS HAS REACHED OR     */
/*                    EXCEEDED 50 * (N + 3).                            */

/*         INFO = 5   APPROXIMATE JACOBIAN MATRIX IS SINGULAR.          */

/*         INFO = 6   ITERATION IS NOT MAKING GOOD PROGRESS.            */

/*         INFO = 7   ITERATION IS DIVERGING.                           */

/*         INFO = 8   ITERATION IS CONVERGING, BUT TOL IS TOO           */
/*                    SMALL, OR THE CONVERGENCE IS VERY SLOW            */
/*                    DUE TO A JACOBIAN SINGULAR NEAR THE OUTPUT        */
/*                    X OR DUE TO BADLY SCALED VARIABLES.               */

/*       WA IS A WORK ARRAY OF LENGTH LWA.                              */

/*       LWA IS A POSITIVE INTEGER INPUT VARIABLE NOT LESS THAN         */
/*         N * (N + 3).                                                 */

/*     SUBPROGRAMS REQUIRED                                             */

/*       USER-SUPPLIED ...... FCN, BRENTM */

/*     **********                                                       */

int brent(int (*fcn)(), int n, double *x, double *fvec, double tol,
           int info, double *wa, int lwa)
{
  /* Initialized data */

  static double zero = 0.0;

  /* Local variables */

  static double emax;
  static int nfev;
  static double ftol, temp;
  static int mopt;
  static double xtol;
  static int i, maxfev;

  /* Function Body */

  info = 0;

  /*     CHECK THE INPUT PARAMETERS FOR ERRORS. */

  if (n <= 0 || tol < zero || lwa < n * (n + 3)) {
    goto L30;
  }

  /*     DETERMINE AN OPTIMAL VALUE FOR MOPT. */

  emax = zero;

  for(i = 0; i < n; i++) {
    temp = log((double) (i + 1)) / ((double) (n + 2 * i + 1));

    if(temp < emax) {
      goto L20;
    }

    mopt = i;
    emax = temp;
  }

 L20:
  
  /*     CALL BRENTM. */

  maxfev = 50 * (n + 3);

  ftol = tol;
  xtol = tol;

  brentm(fcn, n, x, fvec, ftol, xtol, maxfev, mopt, info, nfev,
         &wa[3 * n], n, &wa[0], &wa[n], &wa[2 * n]);

 L30:

  return 0;
}

/*     **********                                                       */

/*     SUBROUTINE BRENTM                                                */

/*     THE PURPOSE OF THIS SUBROUTINE IS TO FIND A ZERO TO              */
/*     A SYSTEM OF N NONLINEAR EQUATIONS IN N VARIABLES BY A            */
/*     METHOD DUE TO R. BRENT.                                          */

/*     THE SUBROUTINE STATEMENT IS                                      */

/*       BRENTM(FCN, N, X, FVEC, FTOL, XTOL, MAXFEV, MOPT,              */
/*              INFO, NFEV, Q, LDQ, SIGMA, WA1, WA2)                    */

/*     WHERE                                                            */

/*       FCN IS THE NAME OF THE USER-SUPPLIED SUBROUTINE WHICH          */
/*         CALCULATES COMPONENTS OF THE FUNCTION. FCN SHOULD BE         */
/*         DECLARED IN AN EXTERNAL STATEMENT IN THE USER CALLING        */
/*         PROGRAM, AND SHOULD BE WRITTEN AS FOLLOWS.                   */

/*         SUBROUTINE FCN(N, X, FVEC, IFLAG)                            */
/*         INT N, IFLAG                                                 */
/*         DOUBLE X(N), FVEC(N)                                         */
/*         ----------                                                   */
/*         CALCULATE THE IFLAG-TH COMPONENT OF THE FUNCTION             */
/*         AND RETURN THIS VALUE IN FVEC(IFLAG).                        */
/*         ----------                                                   */
/*         RETURN                                                       */
/*         END                                                          */

/*         THE VALUE OF IFLAG SHOULD NOT BE CHANGED BY FCN UNLESS       */
/*         THE USER WANTS TO TERMINATE EXECUTION OF BRENTM.             */
/*         IN THIS CASE SET IFLAG TO A NEGATIVE INTEGER.                */

/*       N IS A POSITIVE INTEGER INPUT VARIABLE SET TO THE NUMBER OF    */
/*         EQUATIONS AND VARIABLES.                                     */

/*       X IS AN ARRAY OF LENGTH N. ON INPUT IT MUST CONTAIN            */
/*         AN ESTIMATE TO THE SOLUTION OF THE SYSTEM OF EQUATIONS.      */
/*         ON OUTPUT X CONTAINS THE FINAL ESTIMATE TO THE SOLUTION      */
/*         OF THE SYSTEM OF EQUATIONS.                                  */

/*       FVEC IS AN ARRAY OF LENGTH N. ON OUTPUT IT CONTAINS            */
/*         THE FINAL RESIDUALS.                                         */

/*       FTOL IS A NONNEGATIVE INPUT VARIABLE. CONVERGENCE              */
/*         OCCURS IF ALL RESIDUALS ARE AT MOST FTOL IN MAGNITUDE.       */

/*       XTOL IS A NONNEGATIVE INPUT VARIABLE. CONVERGENCE              */
/*         OCCURS IF THE RELATIVE ERROR BETWEEN TWO SUCCESSIVE          */
/*         ITERATES IS AT MOST XTOL.                                    */

/*       MAXFEV IS A POSITIVE INTEGER INPUT VARIABLE. TERMINATION       */
/*         OCCURS IF THE NUMBER OF FUNCTION EVALUATIONS IS AT           */
/*         LEAST MAXFEV BY THE END OF AN ITERATION. IN BRENTM,          */
/*         A FUNCTION EVALUATION CORRESPONDS TO N CALLS TO FCN.         */

/*       MOPT IS A POSITIVE INTEGER INPUT VARIABLE. MOPT SPECIFIES      */
/*         THE NUMBER OF TIMES THAT THE APPROXIMATE JACOBIAN IS         */
/*         USED DURING EACH ITERATION WHICH EMPLOYS ITERATIVE           */
/*         REFINEMENT. IF MOPT IS 1, NO ITERATIVE REFINEMENT WILL       */
/*         BE DONE. MAXIMUM EFFICIENCY IS USUALLY OBTAINED IF           */
/*         MOPT MAXIMIZES LOG(K + 1)/(N + 2 * K + 1) FOR K = 1, ..., N. */

/*       INFO IS AN INTEGER OUTPUT VARIABLE SET AS FOLLOWS. IF          */
/*         THE USER HAS TERMINATED EXECUTION, INFO WILL BE SET TO       */
/*         THE (NEGATIVE) VALUE OF IFLAG. SEE DESCRIPTION OF FCN.       */
/*         OTHERWISE                                                    */

/*         INFO = 0   IMPROPER INPUT PARAMETERS.                        */

/*         INFO = 1   ALL RESIDUALS ARE AT MOST FTOL IN MAGNITUDE.      */

/*         INFO = 2   RELATIVE ERROR BETWEEN TWO SUCCESSIVE ITERATES    */
/*                    IS AT MOST XTOL.                                  */

/*         INFO = 3   CONDITIONS FOR INFO = 1 AND INFO = 2 BOTH HOLD.   */

/*         INFO = 4   NUMBER OF FUNCTION EVALUATIONS HAS REACHED OR     */
/*                    EXCEEDED MAXFEV.                                  */

/*         INFO = 5   APPROXIMATE JACOBIAN MATRIX IS SINGULAR.          */

/*         INFO = 6   ITERATION IS NOT MAKING GOOD PROGRESS.            */

/*         INFO = 7   ITERATION IS DIVERGING.                           */

/*         INFO = 8   ITERATION IS CONVERGING, BUT XTOL IS TOO          */
/*                    SMALL, OR THE CONVERGENCE IS VERY SLOW            */
/*                    DUE TO A JACOBIAN SINGULAR NEAR THE OUTPUT        */
/*                    X OR DUE TO BADLY SCALED VARIABLES.               */

/*       NFEV IS AN INTEGER OUTPUT VARIABLE SET TO THE NUMBER OF        */
/*         FUNCTION EVALUATIONS USED IN PRODUCING X. IN BRENTM,         */
/*         A FUNCTION EVALUATION CORRESPONDS TO N CALLS TO FCN.         */

/*       Q IS AN N BY N ARRAY. IF JAC DENOTES THE APPROXIMATE           */
/*         JACOBIAN, THEN ON OUTPUT Q IS (A MULTIPLE OF) AN             */
/*         ORTHOGONAL MATRIX SUCH THAT JAC * Q IS A LOWER TRIANGULAR    */
/*         MATRIX. ONLY THE DIAGONAL ELEMENTS OF JAC * Q NEED           */
/*         TO BE STORED, AND THESE CAN BE FOUND IN SIGMA.               */

/*       LDQ IS A POSITIVE INTEGER INPUT VARIABLE NOT LESS THAN N       */
/*         WHICH SPECIFIES THE LEADING DIMENSION OF THE ARRAY Q.        */

/*       SIGMA IS A LINEAR ARRAY OF LENGTH N. ON OUTPUT SIGMA           */
/*         CONTAINS THE DIAGONAL ELEMENTS OF THE MATRIX JAC * Q.        */
/*         SEE DESCRIPTION OF Q.                                        */

/*       WA1 AND WA2 ARE LINEAR WORK ARRAYS OF LENGTH N.                */

/*     SUBPROGRAMS REQUIRED                                             */

/*       USER-SUPPLIED ...... FCN                                       */

/*     **********                                                       */

int brentm(int (*fcn)(), int n, double *x, double *fvec, double ftol,
           double xtol, int maxfev, int mopt, int info, int nfev,
           double *q, int ldq, double *sigma, double *wa1, double *wa2)
{
  /* Initialized data */

  static double zero  =  0.0;
  static double p05   =  0.05;
  static double scale = 10.0;

  /* System generated locals */

  double d;

  /* Local variables */

  static int conv;
  static double temp;
  static int nier6, nier7, nier8;
  static double h;
  static int i, j, k, m, iflag;
  static double delta, difit;
  static int nsing;
  static double fnorm, xnorm, difit1, fnorm1;
  static int nfcall;
  static double epsmch, sknorm, eta, eps, fky, fkz;
  
  /* Function Body */

  /*     WARNING. */

  /*     THIS IS AN IBM CODE. TO RUN THIS CODE ON OTHER MACHINES IT */
  /*     IS NECESSARY TO CHANGE THE VALUE OF THE MACHINE PRECISION  */
  /*     EPSMCH. THE MACHINE PRECISION IS THE SMALLEST FLOATING     */
  /*     POINT NUMBER EPSMCH SUCH THAT                              */

  /*           1 + EPSMCH .GT. 1                                    */

  /*     IN WORKING PRECISION. IF IN DOUBT ABOUT THE VALUE OF       */
  /*     EPSMCH, THEN THE FOLLOWING PROGRAM SEGMENT DETERMINES      */
  /*     EPSMCH ON MOST MACHINES.                                   */

  /*     EPSMCH = 0.5D0                                             */
  /*   1 CONTINUE                                                   */
  /*     IF (1.D0 + EPSMCH .EQ. 1.D0) GO TO 2                       */
  /*     EPSMCH = 0.5D0 * EPSMCH                                    */
  /*     GO TO 1                                                    */
  /*   2 CONTINUE                                                   */
  /*     EPSMCH = 2.D0 * EPSMCH                                     */

  /*     THE IBM DOUBLE PRECISION EPSMCH.                           */

  epsmch = 2.2204460492503131e-16;

  info   = 0;
  iflag  = 0;
  nfev   = 0;
  nfcall = 0;

  /*     CHECK THE INPUT PARAMETERS FOR ERRORS. */
  
  if((n <= 0) || (ftol < zero) || (xtol < zero) || (maxfev <= 0) ||
     (mopt <= 0) || (ldq < n)) {
    goto L220;
  }

  /*     INITIALIZE SOME OF THE VARIABLES. */

  nier6 = - 1;
  nier7 = - 1;
  nier8 =   0;

  fnorm = zero;
  difit = zero;
  xnorm = zero;

  for(i = 0; i < n; i++) {
    xnorm = maximum(xnorm, fabs(x[i]));
  }

  eps   = sqrt(epsmch);
  delta = scale * xnorm;

  if(xnorm == zero) {
    delta = scale;
  }

  /*     ENTER THE PRINCIPAL ITERATION. */

 L20:
  
  /*     TO PRINT THE ITERATES, PLACE WRITE STATEMENTS */
  /*     FOR THE VECTOR X HERE. */

  nsing = n;

  fnorm1 = fnorm;
  difit1 = difit;
  fnorm  = zero;

  /*     COMPUTE THE STEP H FOR THE DIVIDED DIFFERENCE WHICH */
  /*     APPROXIMATES THE K-TH ROW OF THE JACOBIAN MATRIX. */

  h = eps * xnorm;

  if(h == zero) {
    h = eps;
  }

  for(j = 0; j < n; j++) {
    for(i = 0; i < n; i++) {
      q[i + j * ldq] = zero;
    }

    q[j + j * ldq] = h;
    
    wa1[j] = x[j];
  }

  /*     ENTER A SUBITERATION. */

  for(k = 0; k < n; k++) {
    iflag = k;

    (*fcn)(n, wa1, fvec, iflag);

    fky = fvec[k];

    ++nfcall;
    nfev = nfcall / n;

    if(iflag < 0) {
      goto L230;
    }

    fnorm = maximum(fnorm, fabs(fky));

    /*        COMPUTE THE K-TH ROW OF THE JACOBIAN MATRIX. */

    for(j = k; j < n; j++) {
      for(i = 0; i < n; i++) {
        wa2[i] = wa1[i] + q[i + j * ldq];
      }

      (*fcn)(n, wa2, fvec, iflag);

      fkz = fvec[k];

      ++nfcall;
      nfev = nfcall / n;

      if(iflag < 0) {
        goto L230;
      }

      sigma[j] = fkz - fky;
    }

    fvec[k] = fky;

    /*        COMPUTE THE HOUSEHOLDER TRANSFORMATION TO REDUCE THE K-TH ROW */
    /*        OF THE JACOBIAN MATRIX TO A MULTIPLE OF THE K-TH UNIT VECTOR. */

    eta = zero;

    for(i = k; i < n; i++) {
      eta = maximum(eta, fabs(sigma[i]));
    }

    if(eta == zero) {
      goto L150;
    }

    --nsing;

    sknorm = zero;

    for(i = k; i < n; i++) {
      sigma[i] /= eta;
      sknorm   += sigma[i] * sigma[i];
    }

    sknorm = sqrt(sknorm);

    if(sigma[k] < zero) {
      sknorm = - sknorm;
    }

    sigma[k] += sknorm;

    /*        APPLY THE TRANSFORMATION AND COMPUTE THE MATRIX Q. */

    for(i = 0; i < n; i++) {
      wa2[i] = zero;
    }

    for(j = k; j < n; j++) {
      temp = sigma[j];

      for(i = 0; i < n; i++) {
        wa2[i] += temp * q[i + j * ldq];
      }
    }

    for(j = k; j < n; j++) {
      temp = sigma[j] / (sknorm * sigma[k]);

      for(i = 0; i < n; i++) {
        q[i + j * ldq] -= temp * wa2[i];
      }
    }
    
    /*        COMPUTE THE SUBITERATE. */

    sigma[k] = sknorm * eta;

    temp = fky / sigma[k];

    if(h * abs(temp) > delta) {
      d    = delta / h;
      temp = signum(&d, &temp);
    }
    
    for(i = 0; i < n; i++) {
      wa1[i] += temp * q[i + k * ldq];
    }

  L150:

    ;

    /*  ;  */
  }

  /*     COMPUTE THE NORMS OF THE ITERATE AND CORRECTION VECTOR. */

  xnorm = zero;
  difit = zero;

  for(i = 0; i < n; i++) {
    xnorm = maximum(xnorm, fabs(wa1[i]));
    difit = maximum(difit, fabs(x[i] - wa1[i]));

    x[i] = wa1[i];
  }

  /*     UPDATE THE BOUND ON THE CORRECTION VECTOR. */

  delta = maximum(delta, scale * xnorm);

  /*     DETERMINE THE PROGRESS OF THE ITERATION. */

  conv = fnorm < fnorm1 && difit < difit1 && nsing == 0;

  ++nier6;
  ++nier7;
  ++nier8;

  if(conv) {
    nier6 = 0;
  }

  if(fnorm < fnorm1 || difit < difit1) {
    nier7 = 0;
  }

  if(difit > eps * xnorm) {
    nier8 = 0;
  }

  /*     TESTS FOR CONVERGENCE. */

  if(fnorm <= ftol) {
    info = 1;
  }
  
  if(difit <= xtol * xnorm && conv) {
    info = 2;
  }

  if(fnorm <= ftol && info == 2) {
    info = 3;
  }

  if(info != 0) {
    goto L230;
  }

  /*     TESTS FOR TERMINATION. */

  if(nfev >= maxfev) {
    info = 4;
  }

  if(nsing == n) {
    info = 5;
  }

  if(nier6 == 5) {
    info = 6;
  }

  if(nier7 == 3) {
    info = 7;
  }

  if(nier8 == 4) {
    info = 8;
  }

  if(info != 0) {
    goto L230;
  }

  /*     ITERATIVE REFINEMENT IS USED IF THE ITERATION IS CONVERGING. */

  if(! conv || difit > p05 * xnorm || mopt == 1) {
    goto L220;
  }

  /*     START ITERATIVE REFINEMENT. */

  for(m = 1; m < mopt; m++) {
    fnorm1 = fnorm;
    fnorm = zero;

    for(k = 0; k < n; k++) {
      iflag = k;

      (*fcn)(n, wa1, fvec, iflag);

      fky = fvec[k];

      ++nfcall;
      nfev = nfcall / n;

      if(iflag < 0) {
        goto L230;
      }

      fnorm = maximum(fnorm, fabs(fky));

      /*           ITERATIVE REFINEMENT IS TERMINATED IF IT DOES NOT */
      /*           GIVE A REDUCTION OF THE RESIDUALS. */

      if(fnorm < fnorm1) {
        goto L170;
      }

      fnorm = fnorm1;

      goto L220;

    L170:

      temp = fky / sigma[k];

      for(i = 0; i < n; i++) {
        wa1[i] += temp * q[i + k * ldq];
      }
    }

    /*        COMPUTE THE NORMS OF THE ITERATE AND CORRECTION VECTOR. */

    xnorm = zero;
    difit = zero;

    for(i = 0; i < n; i++) {
      xnorm = maximum(xnorm, fabs(wa1[i]));
      difit = maximum(difit, fabs(x[i] - wa1[i]));

      x[i] = wa1[i];
    }

    /*        STOPPING CRITERIA FOR ITERATIVE REFINEMENT. */

    if(fnorm <= ftol) {
      info = 1;
    }

    if(difit <= xtol * xnorm) {
      info = 2;
    }

    if(fnorm <= ftol && info == 2) {
      info = 3;
    }

    if(nfev >= maxfev && info == 0) {
      info = 4;
    }

    if(info != 0) {
      goto L230;
    }
  }

 L220:

  /*     END OF THE ITERATIVE REFINEMENT. */

  goto L20;

  /*     TERMINATION, EITHER NORMAL OR USER IMPOSED. */

 L230:

  if(iflag < 0) {
    info = iflag;
  }

  return 0;
}

