#include <stdio.h>
#include <math.h>
#include <complex.h>
#include "glxf.h"

double csign(double x1, double x2);
double complex csign2(double complex z);

void model_menu(double x1, double x2)
{
  glColor3f(0.0, 0.0, 0.0);

  yeq1 = 0.0;
  yeq2 = 0.0; 

  alpha1 = 0.0;
  alpha2 = 0.0;
}

void uniform_ring(double x1, double x2)
{
  double ring_factor = 10.0;

  Mkapr = ring_factor * parms[P_KAPPA] * (r_E2 - Ri_E2);
  Mkapt = ring_factor * parms[P_KAPPA] * (Ro_E2 - Ri_E2);

  if(rd >= parms[P_Ro]) {
    alpha1 = (Mkapt * x1) / r_E2;
    alpha2 = (Mkapt * x2) / r_E2; 	     
  } else {
    if(rd >= parms[P_Ri]) {      
      alpha1 = (Mkapr * x1) / r_E2;
      alpha2 = (Mkapr * x2) / r_E2; 	
    } else {
      alpha1 = 0.0;
      alpha2 = 0.0; 
    }
  }
}

void chang_refsdal(double x1, double x2)
{
  alpha1 = x1 / r_E2;
  alpha2 = x2 / r_E2;
}

void second_order(double x1, double x2)
{
  alpha1 = x1 * (1.0 / r_E2 + parms[P_Ko] / r_E3) ;
  alpha2 = x2 * (1.0 / r_E2 + parms[P_Ko] / r_E3); 
}

void double_lens(double x1, double x2)
{
  double rs12, xd1, xd2, rd2, A12, A22, D2, yp1, yp2;

  rs12 = (x1 - parms[P_X_1]) * (x1 - parms[P_X_1])
    + (x2 - parms[P_Y_1]) * (x2 - parms[P_Y_1]);
  
  rs12 /= R_E2;

  /*
    
  A12 = 1.0 - parms[P_SIGMA2] - parms[P_GAMMA2] * cos(parms[P_PHI2]);
  A22 = 1.0 - parms[P_SIGMA2] + parms[P_GAMMA2] * cos(parms[P_PHI2]);
  D2  =                       - parms[P_GAMMA2] * sin(parms[P_PHI2]);

  **/
  
  A12 = 1.0 - parms[P_SIGMA2]
    - parms[P_GAMMA2] * cos((M_PI * parms[P_PHI2]) / 180.0);
  A22 = 1.0 - parms[P_SIGMA2]
    + parms[P_GAMMA2] * cos((M_PI * parms[P_PHI2]) / 180.0);
  D2  = - parms[P_GAMMA2] * sin((M_PI * parms[P_PHI2]) / 180.0);
  
  yp1 = A12 * x1 + D2 * x2;
  yp2 = A22 * x2 + D2 * x1; 
  
  xd1 = yp1 - parms[P_M_1] * parms[P_BETA] * (x1 - parms[P_X_1]) / rs12; 
  xd2 = yp2 - parms[P_M_1] * parms[P_BETA] * (x2 - parms[P_Y_1]) / rs12; 

  rd2  = (xd1 - parms[P_X_2]) * (xd1 - parms[P_X_2])
    + (xd2 - parms[P_Y_2]) * (xd2 - parms[P_Y_2]);

  rd2 /= R_E2;

  alpha1 = parms[P_M_1] * (x1 - parms[P_X_1]) / rs12
    + parms[P_M_2] * (xd1 - parms[P_X_2]) / rd2; 
  alpha2 = parms[P_M_1] * (x2 - parms[P_Y_1]) / rs12
    + parms[P_M_2] * (xd2 - parms[P_Y_2]) / rd2;
}

void singular_isothermal_sphere(double x1, double x2) 
{
  double rbeta = pow(r_E, (parms[P_EXP] - 1.0));

  alpha1 = rbeta * x1;
  alpha2 = rbeta * x2;  
}       

void nonsingular_isothermal_sphere(double x1, double x2) 
{
  double s2 = xc2 + r_E2;

  func = sqrt(s2) - xc; 

  alpha1 = (parms[P_KAPPA] * func * x1) / r_E2;
  alpha2 = (parms[P_KAPPA] * func * x2) / r_E2;  
}

void transparent_sphere(double x1, double x2) 
{
  double s2; 

  if(parms[P_C] > rd) {
    s2 = 1.0 - r2 / R_C2;
    
    func = 1.0 - pow(s2, (3.0 / 2.0));
	  
    alpha1 = (func * x1) / r_E2;
    alpha2 = (func * x2) / r_E2;  
  } else {
    alpha1 = x1 / r_E2;
    alpha2 = x2 / r_E2; 
  }
}

void elliptical(double x1, double x2) 
{
  double Epsilon1 = 1.0 - parms[P_EPSILON];
  double Epsilon2 = 1.0 + parms[P_EPSILON];
  double delta    = 1.0 - parms[P_ALPHA];
  double s2 = xc2 + Epsilon1 * x1_E2 + Epsilon2 * x2_E2; 

  alpha1 = (parms[P_KAPPA] * Epsilon1 * x1) / pow(s2, delta);
  alpha2 = (parms[P_KAPPA] * Epsilon2 * x2) / pow(s2, delta);
}

void king(double x1, double x2) 
{
  /* log1p(x) = log(1.0 + x) */
  
  alpha1 = (parms[P_KAPPA] * x1 * log1p(r_E2)) / r_E2;
  alpha2 = (parms[P_KAPPA] * x2 * log1p(r_E2)) / r_E2; 
}

void truncated_king(double x1, double x2) 
{
  /* hypot(x, y) = sqrt(x * x + y * y) */
  
  func  = log1p(r_E2) + r_E2 / (1.0 + xc2); 
  func -= (4.0 * (hypot(1.0, r_E) - 1.0)) / hypot(1.0, xc);

  alpha1 = (parms[P_KAPPA2] * x1 * func) / r_E2; 
  alpha2 = (parms[P_KAPPA2] * x2 * func) / r_E2;
}             

void hubble(double x1, double x2) 
{
  double hubble_factor = 2.0;

  func = log1p(r_E) - r_E / (1.0 + r_E); 
   
  alpha1 = (hubble_factor * parms[P_KAPPA] * x1 * func) / r_E2;
  alpha2 = (hubble_factor * parms[P_KAPPA] * x2 * func) / r_E2;  
}

void de_vaucouleur(double x1, double x2) 
{
  double poly;

  poly  = 720.0 / tau8 + (720.0 * pow(r_E, (1.0 / 4.0))) / tau7;
  poly += pow(r_E, (3.0 / 2.0)) / tau2;
  poly += (6.0 * pow(r_E, (5.0 / 4.0))) / tau3;
  poly += (30.0 * r_E) / tau4;
  poly += (120.0 * pow(r_E, (3.0 / 4.0))) / tau5;
  poly += (360.0 * sqrt(r_E)) / tau6;
  poly  = 7.0 * poly + pow(r_E, (7.0 / 4.0)) / tau;

  func  = poly * exp(- tau * (pow(r_E, (1.0 / 4.0)) - 1.0));
  func -= (5040.0 * Ftau) / tau8;

  alpha1 = - (8.0 * parms[P_KAPPA] * x1 * func) / r_E2;
  alpha2 = - (8.0 * parms[P_KAPPA] * x2 * func) / r_E2;  
}

void spiral(double x1, double x2) 
{
  func = 2.0 * parms[P_KAPPA] 
    * (1.0 - exp(- parms[P_ALPHA2] * r_E) * (parms[P_ALPHA2] * r_E + 1.0));

  alpha1 = (x1 * func) / (parms[P_ALPHA2] * parms[P_ALPHA2] * r_E2);
  alpha2 = (x2 * func) / (parms[P_ALPHA2] * parms[P_ALPHA2] * r_E2);  
}

void multipole(double x1, double x2) 
{
  double dipol1, dipol2;
  double quadp1, quadp2, qdipterm1, qdipterm2;
    
  dif = x22 - x12;

  dipol1 = (parms[P_DX_QP] * dif
            - 2.0 * parms[P_DY_QP] * x1 * x2) / parms[P_E];
  dipol2 = (- parms[P_DY_QP] * dif
            - 2.0 * parms[P_DX_QP] * x1 * x2) / parms[P_E];

  qdipterm1 = (x2 * (3.0 * x12 - x22)) / R_E2;
  qdipterm2 = (x1 * (3.0 * x22 - x12)) / R_E2;
    
  quadp1 = 2.0 * (parms[P_QXX] * qdipterm2 - parms[P_QXY] * qdipterm1);
  quadp2 = 2.0 * (parms[P_QYY] * qdipterm1 - parms[P_QXY] * qdipterm2);  
  
  alpha1 = (parms[P_KAPPA0] * x1) / r_E2 + (dipol1 / r_E4 + quadp1 / r_E6);
  alpha2 = (parms[P_KAPPA0] * x2) / r_E2 + (dipol2 / r_E4 + quadp2 / r_E6); 
}

void rotation_lens(double x1, double x2) 
{
  func = (parms[P_SX] * x2 - parms[P_SY] * x1) / parms[P_E];

  alpha1 = x1 / r_E2 + (2.0 * func * x1) / r_E4
    + parms[P_SY] / (parms[P_E] * r_E2);
  alpha2 = x2 / r_E2 + (2.0 * func * x2) / r_E4
    - parms[P_SX] / (parms[P_E] * r_E2);
}

void Elliptical_SIS_KK(double x1, double x2) 
{
  double Epsilon1 = 1.0 - parms[P_EPSILON3];
  double Epsilon2 = 1.0 + parms[P_EPSILON3];
  double s2 = Epsilon1 * x1_E2 + Epsilon2 * x2_E2; 
  double factor = sqrt(2.0 * parms[P_EPSILON3]);
  double zeta = factor * x2 / sqrt(s2);

  x1 = x1 / parms[P_E];
  x2 = x2 / parms[P_E];
    
  alpha1 = (parms[P_KAPPA] / factor) * atanl((factor * x1) / sqrt(s2));
  alpha2 = (parms[P_KAPPA] / (2.0 * factor)) 
    * logl((1.0 + zeta) / (1.0 - zeta));  
}       

void Elliptic_King(double x1, double x2)
{
  double f = parms[P_EPSILON2], h;
  double kappak = parms[P_KAPPA], x;
  double complex z, z_2, z_3, z_4, alphakingc;
  
  h = sqrt(1.0 - f * f);
  
  x1 = x1 / parms[P_E];
  x2 = x2 / parms[P_E];
  
  x = sqrt(x1 * x1 + f * f * x2 * x2);

  z = x1 + I * x2;
  
  z_2 = z * z;
  z_3 = h * h * xc * xc - f * f * z_2;
  z_4 = h * h * x * x + f * f * z_2;
  
  alphakingc = 2.0 * kappak * csign2(z)
    * (catanl(csqrtl(z_4) / csqrtl(z_3)) / csqrtl(z_3) 
       - catanl(f * csqrtl(z_2) / csqrtl(z_3)) / sqrt(z_3));
  
  alpha1 =   creal(alphakingc);
  alpha2 = - cimag(alphakingc);
}

void Elliptic_NIS(double x1, double x2)
{
  double f = parms[P_EPSILON2], h;
  double kappanis = parms[P_KAPPA], x;
  double complex z, z_2, z_3, z_4, z_5, z_6, z_7, alphanisc;
  
  h = sqrt(1.0 - f * f);
  
  x1 = x1 / parms[P_E];
  x2 = x2 / parms[P_E];
  
  x = sqrt(x1 * x1 + f * f * x2 * x2);
  
  z = x1 + I * x2;
  
  z_2 = z * z;
  z_3 = h * h * xc * xc + f * f * z * z;
  z_4 = z_3 + 2.0 * h * xc * f * csqrtl(z_2);
  z_5 = z_3 + 2.0 * h * h * x * x; 
  z_6 = h * h * x * x + f * f * z * z;
  z_7 = z_5 + 2.0 * h * sqrt(xc * xc + x * x) * csqrtl(z_6);
  
  alphanisc = (kappanis / h) * csign2(z) * clogl(z_7 / z_4);

  alpha1 =   creal(alphanisc);
  alpha2 = - cimag(alphanisc);
}

void Elliptic_Hubble(double x1, double x2)
{
  double f = parms[P_EPSILON2], h;
  double kappah = parms[P_KAPPA], x;
  double complex z, z_2, z_3, z_4, z_5, z_6, z_7, z_8, z_9, alphahubblec;
  
  h = sqrt(1.0 - f * f);
  
  x1 = x1 / parms[P_E];
  x2 = x2 / parms[P_E];
  
  x = sqrt(x1 * x1 + f * f * x2 * x2);
  
  z = x1 + I * x2;
  
  z_2 = z * z;
  z_3 = h * h * xc * xc + f * f * z * z;
  z_4 = h * h * x * x + f * f * z * z;
  z_5 = h * h * xc * x - f * f* z * z;
  z_6 = xc * csqrtl(z_4) - (xc + x) * f * csqrtl(z_2);
  z_7 = ((h * h * xc * xc) / (f * csqrtl(z_2)) 
	 + f * csqrtl(z_2) + csqrtl(z_3)) / xc;
  z_8 = (h * h * (xc + x)) / (z_5/csqrtl(z_3) + csqrtl(z_4));
  z_9 = ((xc + x) * cpowl(z_3, (3.0 / 2.0)));
  
  alphahubblec = (2.0 * kappah * csign2(z) 
		  * (csqrtl(z_3) * z_6 + f * f * (xc + x) * z2 
		     * (clogl(z_7) - clogl(z_8)))) / z_9;
  
  alpha1 =   creal(alphahubblec);
  alpha2 = - cimag(alphahubblec);
}

void Elliptic_Plummer(double x1, double x2)
{
  yeq1 = 0.0;
  yeq2 = 0.0; 

  alpha1 = 0.0;
  alpha2 = 0.0;
}

void Elliptic_KK(double x1, double x2)
{
  /* atanh(z) = [log_e((1 + z) / (1 - z))] / 2 */

  double q1 = parms[P_Q3] * cos((M_PI * parms[P_THETA_INC]) / 180.0); 
  double q2 = sin((M_PI * parms[P_THETA_INC]) / 180.0); 
  double B_I = 0.0, q = 0.0, psi = 0.0, zeta = 0.0, factor = 0.0;

  x1 = x1 / parms[P_E];
  x2 = x2 / parms[P_E];

  q = sqrt(q1 * q1 + q2 * q2);

  factor = sqrt(1.0 - q * q);

  B_I = parms[P_KAPPA] / factor; 
	 
  psi = sqrt(q * q * (xc * xc + x1 * x1) + x2 * x2); 
 
  zeta = factor * x2 / (psi + q * q * xc);

  alpha1 = B_I * atanl(factor * x1 / (psi + xc));
  alpha2 = B_I * (logl((1.0 + zeta) / (1.0 - zeta)) / 2.0);
}

void Navarro_Frenk_White(double x1, double x2)
{
  double func = 0.0, func1 = 0.0, func2 = 0.0;

  if(r_E > xc) {
    func1 = 2.0 / sqrt(r_E2 - xc2);
    func2 = sqrt((r_E - xc) / (r_E + xc));
    func = func1 * atanl(func2);
  } else if(r_E < xc) {
    func1 = 1.0 / sqrt(xc2 - r_E2);
    func2 = sqrt((xc - r_E) / (r_E + xc));
    func = func1 * logl((1.0 + func2)/(1.0 - func));
  } else {
    func = 1.0;
  }

  func += log(r_E / 2.0);

  alpha1 = (2.0 * parms[P_KAPPA] * func * x1) / r_E2;
  alpha2 = (2.0 * parms[P_KAPPA] * func * x2) / r_E2;
}

/*

               /  1    if Re(x) > 0 or Re(x) = 0 and Im(x) > 0
    csign(x) = <
               \ -1    if Re(x) < 0 or Re(x) = 0 and Im(x) < 0

**/

double csign(double x1, double x2) 
{
  double csignum = 1.0;

  if((x1 > 0.0 || x1 == 0.0) && x2 > 0) {
    csignum = 1.0;
  } else if((x1 < 0.0 || x1 == 0.0) && x2 < 0) {
    csignum = - 1.0;
  }

  return(csignum);    
}

double complex csign2(double complex z)
{
  double complex csignum = 0.0;
  
  csignum = csqrtl(z * z) / z;

  return(csignum);    
}

/* Barkana's Fastell */

#include "Fastell_2.c"
#include "Dqnc79_2.c"

void Barkana(double x1, double x2)
{
  /* magmx was [2][2] */
  
  double q, s2, x1in, x2in, gam, defl[2], arat, magmx[4];
  double Magxx, Magxy, Magyx, Magyy; 

  /* Subroutine */ 
  
  extern int fastellmag();
  
  x1in = x1 / parms[P_E];
  x2in = x2 / parms[P_E];
  q    = parms[P_KAPPA];
  gam  = parms[P_ALPHA3];
  arat = parms[P_EPSILON2];
  s2   = xc2;
  
  fastellmag(&x1in, &x2in, &q, &gam, &arat, &s2, defl, magmx);
  
  alpha1 = defl[0];
  alpha2 = defl[1];
  
  /* Jac[0][0] = Magxx, Jac[0][1] = Magxy, 
     Jac[1][0] = Magyx, Jac[1][1] = Magyy */

  Magxx = magmx[0];
  Magxy = magmx[2];
  Magyx = magmx[1];
  Magyy = magmx[3];
} 

void lens_equation(double x1, double x2, double *y1, double *y2)
{
  double d_0 = (M_PI * parms[P_PHI]) / 180.0;
  double alpha_x, alpha_y;

  /*

  x1 = x1 - xl;
  x2 = x2 - yl;

  **/

  x12 = x1 * x1;
  x22 = x2 * x2;
  
  r2 = x12 + x22;
  r4 = r2 * r2;
  r6 = r2 * r4;
  rd = sqrt(r2);     

  x1_E = x1 / parms[P_E];
  x2_E = x2 / parms[P_E];

  x1_E2 = x1_E * x1_E;
  x2_E2 = x2_E * x2_E;
  
  R_E2 = parms[P_E] * parms[P_E];
  R_E3 = parms[P_E] * R_E2;  
  R_E4 = R_E2 * R_E2;
  R_E6 = R_E2 * R_E4;
  
  r_E2 = x1_E2 + x2_E2;
  r_E  = sqrt(r_E2);
  r_E3 = r_E  * r_E2;
  r_E4 = r_E2 * r_E2;
  r_E6 = r_E2 * r_E4;
  
  R_C2 = parms[P_C] * parms[P_C];

  xc  = parms[P_C] / parms[P_E];
  xc2 = xc * xc;

  /*
    
#if 0
  xt  = parms[P_T] / parms[P_C];
  xt2 = xt * xt;
#endif

  **/

  xt  = parms[P_XT] / parms[P_E];
  xt2 = xt * xt;

  Ri_E = parms[P_Ri] / parms[P_E];
  Ro_E = parms[P_Ro] / parms[P_E];

  Ri_E2 = Ri_E * Ri_E;
  Ro_E2 = Ro_E * Ro_E;

  /*
    
  A1 = 1.0 - parms[P_SIGMA] - parms[P_GAMMA] * cos(parms[P_PHI]);
  A2 = 1.0 - parms[P_SIGMA] + parms[P_GAMMA] * cos(parms[P_PHI]);
  D  =                      - parms[P_GAMMA] * sin(parms[P_PHI]);

  A1 = 1.0 - parms[P_SIGMA]
    - parms[P_GAMMA] * cos((M_PI * parms[P_PHI]) / 180.0);
  A2 = 1.0 - parms[P_SIGMA]
    + parms[P_GAMMA] * cos((M_PI * parms[P_PHI]) / 180.0);
  D  = - parms[P_GAMMA] * sin((M_PI * parms[P_PHI]) / 180.0);

  **/

  /*

  A1 = 1.0 - parms[P_SIGMA] - parms[P_GAMMA] * cos(d_0);
  A2 = 1.0 - parms[P_SIGMA] + parms[P_GAMMA] * cos(d_0);
  D  = - parms[P_GAMMA] * sin(d_0);

  **/

  A1 = 1.0 - parms[P_SIGMA] - parms[P_GAMMA] * cos(2.0 * d_0);
  A2 = 1.0 - parms[P_SIGMA] + parms[P_GAMMA] * cos(2.0 * d_0);
  D  = - parms[P_GAMMA] * sin(2.0 * d_0);

  /*
    
  yeq1 = A1 * x1 + D * x2 + xl;
  yeq2 = A2 * x2 + D * x1 + yl; 

  **/

  black_hole = 0;

  // dist_1 = 0.0;
  // dist_2 = 0.0;
  
  sdist2 = ((x1 - dist_1) * (x1 - dist_1) 
	    + (x2 - dist_2) * (x2 - dist_2)) / (parms[P_E] * parms[P_E]);
  
  if(black_hole == 1) {
    alpha_x = (bh_beta * (x1 - dist_1)) / sdist2;
    alpha_y = (bh_beta * (x2 - dist_2)) / sdist2; 
  } else {
    alpha_x = 0.0;
    alpha_y = 0.0; 
  }
  
  yeq1 = A1 * x1 + D * x2 - alpha_x;
  yeq2 = A2 * x2 + D * x1 - alpha_y; 

  models[current_model](x1, x2);
  
  // printf("Current Model = %d \n", current_model);
  // printf("alpha1 = %lf \n", alpha1);
  // printf("alpha2 = %lf \n", alpha2);

  *y1 = yeq1 - alpha1;
  *y2 = yeq2 - alpha2;
}

/* #include "steepest_descent.c" */









